import java.io.*;
import java.net.*;
import org.garret.perst.*;

class Queue extends Persistent { 
    String  phone;
    Message first;
    Message last;
    int     size;

    static final int MAX_QUEUE_LENGTH = 16;

    boolean add(Message msg) { 
        if (size < MAX_QUEUE_LENGTH) { 
            if (first == null) { 
                first = msg;
            } else { 
                last.next = msg;
                last.modify();
            }
            size += 1;
            last = msg;
            modify();                           
            return true;
        } else { 
            return false;
        }
    }

    Message get() { 
        if (first != null) {
            Message msg = first;
            first = msg.next;
            size -= 1;
            modify();
            return msg;
        }
        return null;
    }
}

class Message extends Persistent { 
    Message next;
    String from;
    String body;
}

public class Server extends Thread { 
    static final int DEFAULT_PORT = 8008;
    static final int DEFAULT_BIND_QUEUE_LEN = 50;
    static final int MAX_REQUEST_LENGTH = 64*1024;
    static final int MAX_CMD_LENGTH = 1024;
    
    volatile boolean running;
    boolean     trace;
    FieldIndex  index;
    int         port;
    int         queueLen;
    PrintStream log;
    Storage     db;
    
    public static void main(String[] args) throws Exception { 
        String portStr = System.getProperty("port");
        String traceStr = System.getProperty("trace");
        String queueStr = System.getProperty("queue");
        boolean trace = traceStr != null ? Boolean.getBoolean(traceStr) : true;
        int port = portStr != null ? Integer.parseInt(portStr) : DEFAULT_PORT;
        int queue = queueStr != null ? Integer.parseInt(queueStr) : DEFAULT_BIND_QUEUE_LEN;
        PrintStream log = args.length > 0 ? new PrintStream(new FileOutputStream(args[0])) : System.err;
        Server server = new Server(port, queue, trace, log);
        if (args.length == 0) { 
            byte[] buf = new byte[MAX_CMD_LENGTH];
            int rc;
            System.out.print("> ");
            while ((rc = System.in.read(buf)) >= 0) { 
                String cmd = new String(buf, 0, rc).trim().toLowerCase();
                if (cmd.equals("quit") || cmd.equals("exit")) {
                    break;
                } else if (cmd.startsWith("add")) { 
                    if (!server.addQueue(cmd.substring(3).trim())) { 
                        log.println("Phone is alreay registred");
                    }
                } else { 
                    log.println("Unknown command " + cmd);
                }
                System.out.print("> ");
            }
            System.out.println("Terminating...");
            server.terminate();
        } else {
            server.join();
        }
    }

    public Server(int port, int queueLen, boolean trace, PrintStream log) { 
        this.port = port;
        this.queueLen = queueLen;
        this.trace = trace;
        this.log = log;
        running = true;
        start();
    }

    static int locateEmptyLine(byte[] buf, int offs, int len) { 
        for (int i = offs; i < len-1; i++) { 
            if (buf[i] == '\n' && (buf[i+1] == '\n' || (buf[i+1] == '\r' && i+2 < len && buf[i+2] == '\n'))) { 
                return i;
            }
        }
        return -1;
    }
    
    static String getHeader(String request, String hdr) {
        int i = request.indexOf(hdr);
        if (i >= 0) { 
            int j = request.indexOf('\n', i);
            return ((j < 0) ? request.substring(i + hdr.length()) : request.substring(i + hdr.length(), j)).trim();
        }
        return null;
    }
    
    class Client extends Thread {
        Socket socket;

        Client(Socket socket) { 
            this.socket = socket;
            setDaemon(true);
            start();
        }
        
                
        public void run() { 
            try { 
                InputStream in = socket.getInputStream();
                OutputStream out = socket.getOutputStream();
                byte[] buf = new byte[MAX_REQUEST_LENGTH];
                int offs = 0;
                int rc;
                int hdrEnd = 0; 
                String hdr = null;
                String from = null;
                while ((rc = in.read(buf, offs, buf.length-offs)) >= 0) { 
                    offs += rc;
                    int eol = locateEmptyLine(buf, hdrEnd, offs);
                    if (eol > 0) { 
                        if (hdr == null) { 
                            while (eol < offs && (buf[eol] == '\n' || buf[eol] == '\r')) { 
                                eol += 1;
                            }                                                                        
                            hdrEnd = eol;
                            hdr = new String(buf, 0, hdrEnd);
                            from = getHeader(hdr, "From:");
                            if (trace) { 
                                log.println("Receive request from " + from + ":");
                                log.println(hdr);
                                log.println("-------------------------------");
                            }
                            if (from == null) { 
                                log.println("Invalid request " + hdr);
                                break;
                            }
                            if (hdr.startsWith("GET")) { 
                                Message msg = getMessage(from);
                                if (msg == null) { 
                                    out.write("HTTP/1.1 404 NOT FOUND\n\n".getBytes());
                                } else {
                                    out.write(("HTTP/1.1 200 OK\nFrom: " 
                                               + msg.from + "\nContent-Length: " + msg.body.length() 
                                               + "\n\n" + msg.body).getBytes());
                                }
                                break;
                            } else if (!hdr.startsWith("POST")) {
                                log.println("Invalid request " + hdr);
                                break;
                            } 
                            if (locateEmptyLine(buf, hdrEnd, offs) < 0) { 
                                continue;
                            }
                        } 
                        String body = new String(buf, hdrEnd, offs-hdrEnd);
                        String to = getHeader(hdr, "To:");
                        if (trace) { 
                            log.println("Message to " + to + ": ");
                            log.println(body);
                            log.println("-------------------------------");
                        }
                        if (to != null && body.length() != 0) {
                            if (sendMessage(from, to, body)) { 
                                out.write("HTTP/1.1 200 OK\n\n".getBytes());
                            } else { 
                                log.println("Destination phone " + to + " is not registered");
                                out.write("HTTP/1.1 403 FORBIDDEN\n\n".getBytes());
                            }
                        } else { 
                            log.println("Invalid address " + to + " or body " + body);
                        }
                        break;
                    }
                }
                in.close();
                out.close();
                socket.close();
            } catch (Exception x) { 
                x.printStackTrace();
            }
        }
    }

    public boolean addQueue(String phone) { 
        index.exclusiveLock();
        Queue queue = new Queue();
        queue.phone = phone;          
        boolean result = index.put(queue);
        db.commit();
        index.unlock();
        return result;
    }

    public boolean sendMessage(String from, String to, String body) { 
        index.exclusiveLock();
        Queue queue = (Queue)index.get(to);
        boolean result = false;
        if (queue != null) {
            Message msg = new Message();
            msg.from = from;
            msg.body = body;
            result = queue.add(msg);
        }
        db.commit();
        index.unlock();
        return result;
    }

    public Message getMessage(String phone) { 
        index.sharedLock();
        Queue queue = (Queue)index.get(phone);
        Message msg = null;
        if (queue != null) {
            msg = queue.get();
        } else { 
            index.unlock();
            index.exclusiveLock();
            queue = (Queue)index.get(phone);
            if (queue == null) { 
                queue = new Queue();
                queue.phone = phone;          
                index.put(queue);
                db.commit();
            } else { 
                msg = queue.get();
            }
        }
        index.unlock();
        return msg;
    }

    public void run() { 
        ServerSocket gate;
        try { 
            gate = new ServerSocket(); 
            gate.setReuseAddress(true);
            gate.bind(new InetSocketAddress((InetAddress)null, port), queueLen);
        } catch (IOException x) { 
            x.printStackTrace();
            return;
        }
        db = StorageFactory.getInstance().createStorage();
        db.open("server.dbs");        
        index = (FieldIndex)db.getRoot();
        if (index == null) { 
            index = db.createFieldIndex(Queue.class, "phone", false);
            db.setRoot(index);
        }
        while (running) { 
            try { 
                Socket socket = gate.accept();
                new Client(socket);
            } catch (Throwable x) {
                x.printStackTrace();
            }               
        }
        db.close();
    }

    public void terminate() throws Exception { 
        running = false;
        interrupt();
        try {
             Socket dummy = new Socket("localhost", port);
             dummy.close();
        } catch (IOException x) {}
        join();
    }
}
            
            
